from __future__ import annotations

from typing import TYPE_CHECKING

from avilla.core.exceptions import permission_error_message
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.qqapi.const import PRIVILEGE_TRANS
from avilla.standard.core.privilege import MuteAllCapability, Privilege
from avilla.standard.core.profile import Summary, SummaryCapability

if TYPE_CHECKING:
    from avilla.qqapi.account import QQAPIAccount  # noqa
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIChannelActionPerform((m := AccountCollector["QQAPIProtocol", "QQAPIAccount"]())._):
    m.namespace = "avilla.protocol/qqapi::action"
    m.identify = "channel"

    @m.pull("land.guild.channel", Summary)
    async def get_summary(self, target: Selector, route: ...) -> Summary:
        result = await self.account.connection.call_http("get", f"channels/{target.pattern['channel']}", {})
        return Summary(result["name"], None)

    @m.pull("land.guild.channel", Privilege)
    async def get_privilege(self, target: Selector, route: ...) -> Privilege:
        result = await self.account.connection.call_http("get", f"channels/{target.pattern['channel']}", {})
        value = int(result["permissions"]) & 1 == 1
        return Privilege(value, value)

    @m.pull("land.guild.channel", Privilege >> Summary)
    async def get_privilege_summary(self, target: Selector, route: ...) -> Summary:
        result = await self.account.connection.call_http("get", f"channels/{target.pattern['channel']}", {})
        return Summary(PRIVILEGE_TRANS[result["permissions"]], "permission of account self")

    @SummaryCapability.set_name.collect(m, target="land.guild.channel", route=Summary)
    async def set_name(self, target: Selector, route: ..., name: str):
        result = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}/api_permission", {})
        for api in result["apis"]:
            if api["path"] == "channels/{channel_id}" and api["method"] == "PATCH":
                await self.account.connection.call_http(
                    "patch", f"channels/{target.pattern['channel']}", {"name": name}
                )
                return
        raise PermissionError(permission_error_message(f"set_name@{target.path}", "read", ["manage"]))

    @MuteAllCapability.mute_all.collect(m, target="land.guild.channel")
    async def channel_mute_all(self, target: Selector, route: ...):
        await self.account.connection.call_http(
            "patch", f"channels/{target.pattern['channel']}", {"speak_permission": 2}
        )

    @MuteAllCapability.unmute_all.collect(m, target="land.guild.channel")
    async def channel_unmute_all(self, target: Selector, route: ...):
        await self.account.connection.call_http(
            "patch", f"channels/{target.pattern['channel']}", {"speak_permission": 1}
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from avilla.core.event import AvillaEvent

from .metadata import Audit

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface


@dataclass
class MessageAudited(AvillaEvent):
    audit: Audit

    class Dispatcher(AvillaEvent.Dispatcher):
        @classmethod
        async def catch(cls, interface: DispatcherInterface[MessageAudited]):
            if interface.annotation is Audit:
                return interface.event.audit
            return await AvillaEvent.Dispatcher.catch(interface)


@dataclass
class MessageAuditPass(MessageAudited):
    pass


@dataclass
class MessageAuditReject(MessageAudited):
    pass



from __future__ import annotations

from avilla.core.ryanvk import Fn, TargetOverload
from avilla.core.selector import Selector
from graia.ryanvk.capability import Capability


class RoleCreate(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def create(
        self, target: Selector, name: str, hoist: bool | None = None, color: int | None = None
    ) -> Selector:
        ...


class RoleDelete(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def delete(self, target: Selector) -> None:
        ...


class RoleEdit(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def edit(
        self, target: Selector, name: str | None = None, hoist: bool | None = None, color: int | None = None
    ) -> None:
        ...


class RoleMemberCapability(Capability):
    @Fn.complex({TargetOverload(): ["target"]})
    async def add(self, target: Selector, member: Selector) -> None:
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def remove(self, target: Selector, member: Selector) -> None:
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

from avilla.core.elements import Face, Notice, NoticeAll, Picture, Text
from avilla.core.ryanvk.collector.application import ApplicationCollector
from avilla.core.selector import Selector
from avilla.qqapi.capability import QQAPICapability
from avilla.qqapi.element import Ark, ArkKv, Embed, Reference
from avilla.qqapi.resource import QQAPIImageResource
from graia.ryanvk import OptionalAccess

if TYPE_CHECKING:
    from avilla.core.context import Context


class QQAPIMessageDeserializePerform((m := ApplicationCollector())._):
    m.namespace = "avilla.protocol/qqapi::message"
    m.identify = "deserialize"

    # LINK: https://github.com/microsoft/pyright/issues/5409
    context: OptionalAccess[Context] = OptionalAccess()

    @m.entity(QQAPICapability.deserialize_element, raw_element="text")
    async def text(self, raw_element: dict) -> Text:
        return Text(raw_element["text"])

    @m.entity(QQAPICapability.deserialize_element, raw_element="emoji")
    async def emoji(self, raw_element: dict) -> Face:
        return Face(raw_element["id"])

    @m.entity(QQAPICapability.deserialize_element, raw_element="attachment")
    async def attachment(self, raw_element: dict) -> Picture:
        resource = QQAPIImageResource(Selector().land("qqguild").picture(url := raw_element["url"]), url)
        return Picture(resource)

    @m.entity(QQAPICapability.deserialize_element, raw_element="mention_user")
    async def mention(self, raw_element: dict) -> Notice:
        if self.context:
            return Notice(self.context.scene.member(raw_element["user_id"]))
        return Notice(Selector().land("qqguild").member(raw_element["user_id"]))

    @m.entity(QQAPICapability.deserialize_element, raw_element="mention_channel")
    async def mention(self, raw_element: dict) -> Notice:
        if self.context:
            return Notice(
                Selector().land("qqguild").guild(self.context.scene["guild"]).channel(raw_element["channel_id"])
            )
        return Notice(Selector().land("qqguild").channel(raw_element["channel_id"]))

    @m.entity(QQAPICapability.deserialize_element, raw_element="mention_everyone")
    async def mention_everyone(self, raw_element: dict) -> NoticeAll:
        return NoticeAll()

    @m.entity(QQAPICapability.deserialize_element, raw_element="message_reference")
    async def message_reference(self, raw_element: dict) -> Reference:
        return Reference(raw_element["message_id"])

    @m.entity(QQAPICapability.deserialize_element, raw_element="embed")
    async def embed(self, raw_element: dict) -> Embed:
        return Embed(
            raw_element["title"],
            raw_element.get("prompt"),
            raw_element.get("thumbnail", {})["url"],
            [i["name"] for i in raw_element.get("fields", [])],
        )

    @m.entity(QQAPICapability.deserialize_element, raw_element="ark")
    async def ark(self, raw_element: dict) -> Ark:
        kvs: list[ArkKv] = []
        for i in raw_element.get("kv", []):
            if i.get("value"):
                kvs.append(ArkKv(**i))
            else:
                obj: list[dict] = [{slot["key"]: slot["value"] for slot in kv["obj_kv"]} for kv in i["obj"]]
                kvs.append(ArkKv(i["key"], obj=obj))
        return Ark(raw_element["template_id"], kvs)



from __future__ import annotations

from typing import TYPE_CHECKING

from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.standard.core.common import Count
from avilla.standard.core.profile import Summary

if TYPE_CHECKING:
    from avilla.qqapi.account import QQAPIAccount  # noqa
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIGuildActionPerform((m := AccountCollector["QQAPIProtocol", "QQAPIAccount"]())._):
    m.namespace = "avilla.protocol/qqapi::action"
    m.identify = "guild"

    @m.pull("land.guild", Summary)
    async def get_summary(self, target: Selector, route: ...) -> Summary:
        result = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}", {})
        return Summary(result["name"], result["description"])

    @m.pull("land.guild", Count)
    async def get_count(self, target: Selector, route: ...) -> Count:
        result = await self.account.connection.call_http("get", f"guilds/{target.pattern['guild']}", {})
        return Count(result["member_count"], result["max_members"])



from __future__ import annotations

from dataclasses import dataclass

from avilla.core.elements import Element


@dataclass
class Embed(Element):
    title: str
    prompt: str | None = None
    thumbnail: str | None = None
    fields: list[str] | None = None

    def __str__(self):
        return "[$Embed]"

    def __repr__(self):
        return f"[$Embed:title={self.title};prompt={self.prompt};thumbnail={self.thumbnail};fields={self.fields}]"


@dataclass
class ArkKv:
    key: str
    value: str | None = None
    obj: list[dict] | None = None


@dataclass
class Ark(Element):
    template_id: int | None = None
    kv: list[ArkKv] | None = None

    def __str__(self):
        return "[$Ark]"

    def __repr__(self):
        return f"[$Ark:template_id={self.template_id};kv={self.kv}]"


@dataclass
class Reference(Element):
    message_id: str
    ignore_get_message_error: bool = False

    def __str__(self):
        return "[$Reference:id={self.message_id}]"

    def __repr__(self):
        return f"[$Reference:id={self.message_id};ignore_error={self.ignore_get_message_error}]"


@dataclass
class Markdown(Element):
    template_id: int | None = None
    content: str | None = None
    custom_template_id: int | None = None
    params: dict[str, list[str]] | None = None

    def __str__(self):
        return "[$Markdown]"

    def __repr__(self):
        return f"[$Markdown:content={self.content}]"


@dataclass
class ActionPermission:
    type: int | None = None
    specify_role_ids: list[str] | None = None
    specify_user_ids: list[str] | None = None


@dataclass
class Action:
    type: int | None = None
    permission: ActionPermission | None = None
    click_limit: int | None = None
    unsupport_tips: str | None = None
    data: str | None = None
    at_bot_show_channel_list: bool | None = None


@dataclass
class RenderData:
    label: str | None = None
    visited_label: str | None = None
    style: int | None = None


@dataclass
class Button:
    id: str | None = None
    render_data: RenderData | None = None
    action: Action | None = None


@dataclass
class Keyboard(Element):
    id: str | None = None
    content: list[list[Button]] | None = None

    def __str__(self):
        return "[$Keyboard]"

    def __repr__(self):
        return f"[$Keyboard:id={self.id};content={self.content}]"



from __future__ import annotations

from avilla.core.resource import Resource
from avilla.core.selector import Selector


class QQAPIResource(Resource[bytes]):
    url: str

    def __init__(self, selector: Selector, url: str):
        super().__init__(selector)
        self.url = url


class QQAPIImageResource(QQAPIResource):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from aiohttp import ClientSession

from avilla.core.builtins.capability import CoreCapability
from avilla.core.ryanvk.collector.protocol import ProtocolCollector
from avilla.qqapi.resource import QQAPIImageResource, QQAPIResource

if TYPE_CHECKING:
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIResourceFetchPerform((m := ProtocolCollector["QQAPIProtocol"]())._):
    m.namespace = "avilla.protocol/qqapi::resource_fetch"

    @m.entity(CoreCapability.fetch, resource=QQAPIResource)
    @m.entity(CoreCapability.fetch, resource=QQAPIImageResource)
    async def fetch_resource(self, resource: QQAPIResource) -> bytes:
        async with ClientSession() as session:
            async with session.get(resource.url) as resp:
                return await resp.read()



from __future__ import annotations

from typing import Any

from graia.amnesia.message import Element, MessageChain

from avilla.core.event import AvillaEvent
from avilla.core.ryanvk.collector.application import ApplicationCollector
from avilla.standard.core.application.event import AvillaLifecycleEvent
from graia.ryanvk import Fn, PredicateOverload, SimpleOverload, TypeOverload

from .utils import handle_text


class QQAPICapability((m := ApplicationCollector())._):
    @Fn.complex({SimpleOverload(): ["event_type"]})
    async def event_callback(self, event_type: str, raw_event: dict) -> AvillaEvent | AvillaLifecycleEvent | None:
        ...

    @Fn.complex({PredicateOverload(lambda _, raw: raw["type"]): ["raw_element"]})
    async def deserialize_element(self, raw_element: dict) -> Element:
        ...

    @Fn.complex({TypeOverload(): ["element"]})
    async def serialize_element(self, element: Any) -> str | tuple[str, Any]:
        ...

    async def deserialize(self, event: dict):
        elements = []

        if message_reference := event.get("message_reference"):
            elements.append(await self.deserialize_element({"type": "message_reference", **message_reference}))
        if event.get("mention_everyone", False):
            elements.append(await self.deserialize_element({"type": "mention_everyone"}))
        if "content" in event:
            for i in handle_text(event["content"]):
                elements.append(await self.deserialize_element(i))
        if attachments := event.get("attachments"):
            for i in attachments:
                elements.append(await self.deserialize_element({"type": "attachment", "url": i["url"]}))
        if embeds := event.get("embeds"):
            for i in embeds:
                elements.append(await self.deserialize_element({"type": "embed", **i}))
        if ark := event.get("ark"):
            elements.append(await self.deserialize_element({"type": "ark", **ark}))

        return MessageChain(elements)

    async def serialize(self, message: MessageChain):
        res = {}
        content = ""

        # def pro_serialize(message: list[dict]):
        #     res = {}
        #     content = ""
        #     for elem in message:
        #         if elem["type"] == "mention_everyone":
        #             content += "@everyone"
        #         elif elem["type"] == "mention_user":
        #             content += f"<@{elem['user_id']}>"
        #         elif elem["type"] == "mention_channel":
        #             content += f"<#{elem['channel_id']}"
        #         elif elem["type"] == "emoji":
        #             content += f"<emoji:{elem['id']}>"
        #         elif elem["type"] == "text":
        #             content += escape(elem["text"])
        #         elif elem["type"] == "attachment":
        #             res["image"] = elem["url"]
        #         elif elem["type"] == "local_iamge":
        #             res["file_image"] = elem["content"]
        #         elif elem["type"] == "markdown":
        #             res["markdown"] = elem["content"]
        #             res["markdown"].pop("type")
        #         elif elem["type"] == "keyboard":
        #             res["keyboard"] = elem["content"]
        #             res["keyboard"].pop("type")
        #         elif elem["type"] == "embed":
        #             res["embed"] = elem
        #             res["embed"].pop("type")
        #         elif elem["type"] == "ark":
        #             res["ark"] = elem
        #             res["ark"].pop("type")
        #         elif elem["type"] == "message_reference":
        #             res["message_reference"] = elem
        #             res["message_reference"].pop("type")
        #     if content:
        #         res["content"] = content
        #     return res

        for element in message:
            elem = await self.serialize_element(element)
            if isinstance(elem, str):
                content += elem
            else:
                res[elem[0]] = elem[1]
        if content:
            res["content"] = content
        return res

    async def handle_event(self, event: dict):
        maybe_event = await self.event_callback(event)

        if maybe_event is not None:
            self.avilla.broadcast.postEvent(maybe_event)



from __future__ import annotations

from typing import TYPE_CHECKING

from avilla.core.builtins.capability import CoreCapability
from avilla.core.context import Context
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector

if TYPE_CHECKING:
    from avilla.qqapi.account import QQAPIAccount  # noqa
    from avilla.qqapi.protocol import QQAPIProtocol  # noqa


class QQAPIContextPerform((m := AccountCollector["QQAPIProtocol", "QQAPIAccount"]())._):
    m.namespace = "avilla.protocol/qqapi::context"

    @m.entity(CoreCapability.get_context, target="land.guild.channel")
    def get_context_from_channel(self, target: Selector, *, via: Selector | None = None):
        return Context(
            self.account,
            target,
            target,
            target,
            target.member(self.account.route["account"]),
        )

    @m.entity(CoreCapability.get_context, target="land.guild.channel.member")
    def get_context_from_guild_member(self, target: Selector, *, via: Selector | None = None):
        return Context(
            self.account,
            target,
            target.into("::guild.channel"),
            target.into("::guild.channel"),
            target.into(f"~.member({self.account.route['account']})"),
        )

    @m.entity(CoreCapability.get_context, target="land.guild.user")
    def get_context_from_user(self, target: Selector, *, via: Selector | None = None):
        if via:
            return Context(
                self.account,
                via,
                target,
                target,
                self.account.route,
            )
        return Context(self.account, target, self.account.route, target, self.account.route)

    @m.entity(CoreCapability.get_context, target="land.group")
    def get_context_from_group(self, target: Selector, *, via: Selector | None = None):
        return Context(
            self.account,
            target,
            target,
            target,
            target.member(self.account.route["account"]),
        )

    @m.entity(CoreCapability.get_context, target="land.friend")
    def get_context_from_friend(self, target: Selector, *, via: Selector | None = None):
        if via:
            return Context(
                self.account,
                via,
                target,
                target,
                self.account.route,
            )
        return Context(self.account, target, self.account.route, target, self.account.route)

    @m.entity(CoreCapability.get_context, target="land.group.member")
    def get_context_from_group_member(self, target: Selector, *, via: Selector | None = None):
        return Context(
            self.account,
            target,
            target.into("::group"),
            target.into("::group"),
            target.into(f"~.member({self.account.route['account']})"),
        )



from __future__ import annotations

from avilla.core.context import Context
from avilla.core.event import MetadataModified, ModifyDetail
from avilla.core.selector import Selector
from avilla.qqapi.capability import QQAPICapability
from avilla.qqapi.collector.connection import ConnectionCollector
from avilla.standard.core.profile import Summary


class QQAPIEventMetadataPerform((m := ConnectionCollector())._):
    m.namespace = "avilla.protocol/qqapi::event"
    m.identify = "metadata"

    @m.entity(QQAPICapability.event_callback, event_type="guild_update")
    async def guild_update(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(str(self.connection.account_id))
        account = self.protocol.avilla.accounts[account_route].account
        land = Selector().land("qq")
        guild = land.guild(str(raw_event["id"]))
        operator = guild.user(str(raw_event["op_user_id"]))
        context = Context(
            account,
            operator,
            guild,
            guild,
            guild.user(account_route["account"]),
        )
        return MetadataModified(
            context,
            guild,
            Summary,
            {
                Summary.inh(lambda x: x.name): ModifyDetail("set", raw_event["name"], None),
                Summary.inh(lambda x: x.description): ModifyDetail("set", raw_event["description"], None),
            },
        )

    @m.entity(QQAPICapability.event_callback, event_type="channel_update")
    async def channel_update(self, event_type: ..., raw_event: dict):
        account_route = Selector().land("qq").account(str(self.connection.account_id))
        account = self.protocol.avilla.accounts[account_route].account
        land = Selector().land("qq")
        guild = land.guild(str(raw_event["guild_id"]))
        channel = guild.channel(str(raw_event["id"]))
        operator = channel.member(str(raw_event["op_user_id"]))
        context = Context(
            account,
            operator,
            channel,
            guild,
            channel.member(account_route["account"]),
        )
        return MetadataModified(
            context, channel, Summary, {Summary.inh(lambda x: x.name): ModifyDetail("set", raw_event["name"], None)}
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from avilla.core.account import AccountStatus, BaseAccount
from avilla.core.selector import Selector
from avilla.qqapi.connection.base import QQAPINetworking

if TYPE_CHECKING:
    from .protocol import QQAPIProtocol


@dataclass
class QQAPIAccount(BaseAccount):
    protocol: QQAPIProtocol
    connection: QQAPINetworking

    def __init__(self, route: Selector, protocol: QQAPIProtocol):
        super().__init__(route, protocol.avilla)
        self.protocol = protocol
        self.status = AccountStatus()

    @property
    def available(self) -> bool:
        return self.connection.alive



from typing import Optional

from avilla.core.exceptions import ActionFailed as BaseActionFailed
from avilla.core.exceptions import HttpRequestError, InvalidAuthentication
from avilla.core.exceptions import NetworkError as BaseNetworkError
from avilla.qqapi.audit import audit_result


class ActionFailed(HttpRequestError, BaseActionFailed):
    def __init__(self, status: int, body: Optional[dict] = None):
        self.body = body or {}
        self.code: Optional[int] = self.body.get("code", None)
        self.message: Optional[str] = self.body.get("message", None)
        self.data: Optional[dict] = self.body.get("data", None)
        super().__init__(status, self.message or "Unknown")

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}: {self.status}, code={self.code}, "
            f"message={self.message}, data={self.data}>"
        )


class UnauthorizedException(ActionFailed, InvalidAuthentication):
    pass


class RateLimitException(ActionFailed):
    pass


class ApiNotAvailable(ActionFailed):
    pass


class NetworkError(BaseNetworkError):
    def __init__(self, msg: Optional[str] = None):
        super().__init__()
        self.msg: Optional[str] = msg
        """错误原因"""

    def __repr__(self):
        return f"<NetWorkError message={self.msg}>"

    def __str__(self):
        return self.__repr__()


class AuditException(Exception):
    """消息审核"""

    def __init__(self, audit_id: str):
        super().__init__()
        self.audit_id = audit_id

    async def get_audit_result(self, timeout: Optional[float] = None):
        """获取审核结果"""
        return await audit_result.fetch(self.audit_id, timeout or 60)
